/**
 * Copyright 2007 Nathanael Cottin
 */
package test;

import java.util.Collection;

import sds.DistributedServerObject;
import sds.data.Message;
import sds.data.Reference;
import sds.error.SdsException;

public final class FeedbackServer extends DistributedServerObject {

	private Message					initMsg				= null;
	private Collection<Reference>	remainingNeighbours	= null;
	private Reference				parent				= null;
	private String					value				= null;

	public FeedbackServer() {
		super(false);
	}
	
	@Override
	public void execute() {
	}

	@Override
	public void processError(SdsException error) {
		System.out.println("Feedback server: " + getReference());
		error.printStackTrace();
	}

	@SuppressWarnings("unchecked")
	@Override
	public boolean receive(Message msg) throws SdsException {
		switch (msg.getTag()) {
			case FeedbackTags.NEIGHBOURS_TAG: {
				Collection<Reference> n = (Collection<Reference>) msg.getData();
				if (n != null) {
					remainingNeighbours = n;
				}

				Reference ref = getReference();
				System.out.println(ref + " - received " + n.size()
					+ " neighbour(s):");
				for (Reference r : n) {
					System.out.println("  " + ref + " -> " + r);
				}

				Message toSend = msg.createResponse();
				toSend.setTag(FeedbackTags.NEIGHBOURS_TAG);
				deliver(toSend, false, getManager());
				// getManager() is equal to msg.getIssuer()
				// in this particular case
				return true;
			}
			case FeedbackTags.INIT_TAG: {
				// Save message (to keep transactional information to
				// notify the end of the algorithm to the feedback manager)
				// Please refer to 'RETURN_TAG' case for 'initMsg' use
				initMsg = msg;
				value = (String) msg.getData();
				System.out.println(getReference() + " - initiated \"" + value
					+ "\"");
				Message toSend = new Message();
				toSend.setTag(FeedbackTags.MESSAGE_TAG);
				toSend.setData(value);
				deliver(toSend, false, remainingNeighbours);
				return true;
			}
			case FeedbackTags.MESSAGE_TAG: {
				Message toSend = msg.createResponse();
				synchronized (this) {
					remainingNeighbours.remove(msg.getIssuer());
					if (value != null) {
						toSend.setTag(FeedbackTags.RETURN_TAG);
						deliver(toSend, false, msg.getIssuer());
						return true;
					}

					// Receiving value for the first time
					value = (String) msg.getData();
				}

				System.out.println(getReference() + " - received \"" + value
					+ "\"");
				parent = msg.getIssuer();
				if (remainingNeighbours.isEmpty()) {
					toSend.setTag(FeedbackTags.RETURN_TAG);
					deliver(msg, false, parent);
					return true;
				}

				toSend.setTag(FeedbackTags.MESSAGE_TAG);
				toSend.setData(value);
				deliver(toSend, false, remainingNeighbours);
				return true;
			}
			case FeedbackTags.RETURN_TAG: {
				boolean empty;
				synchronized (this) {
					remainingNeighbours.remove(msg.getIssuer());
					empty = remainingNeighbours.isEmpty();
				}

				if (empty) {
					if (parent == null) {
						// Respond to feedback manager's initial message
						Message toSend = initMsg.createResponse();
						toSend.setTag(FeedbackTags.END_TAG);
						deliver(toSend, false, getManager());
					}
					else {
						Message toSend = msg.createResponse();
						toSend.setTag(FeedbackTags.RETURN_TAG);
						deliver(toSend, false, parent);
					}
				}

				return true;
			}
			default:
				return false;
		}
	}

	@Override
	public void terminateServer() {
	}
}
